"""
This file is part of jiractl

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import urlparse

from jiractl.commands import base


class JiraLinkMixin(object):
    columns = ("id", "type", "text", "details", "icon")

    TYPE_REMOTE_LINK = 'link'

    def get_parser(self, prog_name):
        parser = super(JiraLinkMixin, self).get_parser(prog_name)
        parser.add_argument("--issue", type=base.utf8, help="Issue ID", required=True)
        return parser

    @staticmethod
    def format_issue_link(link):
        """Converts issue link to tuple."""
        try:
            link_target = link.outwardIssue
            link_type = link.type.outward
        except AttributeError:
            link_target = link.inwardIssue
            link_type = link.type.inward

        return (
            "I{0}".format(link.id), link_type, link_target.fields.summary, link_target.fields.status.name, ""
        )

    def format_remote_link(self, link):
        """Converts issue link to tuple."""
        try:
            icon = link.object.icon.url16x16
        except AttributeError:
            icon = ""

        return "L{0}".format(link.id), self.TYPE_REMOTE_LINK, link.object.title, link.object.url, icon


class ListLinks(JiraLinkMixin, base.JiraList):
    """Gets all links for issue."""

    def take_action(self, parsed_args):
        """Get issue by id."""
        issue_links = [self.format_issue_link(x) for x in self.app.jira.issue(parsed_args.issue).fields.issuelinks]
        remote_links = [self.format_remote_link(x) for x in self.app.jira.remote_links(parsed_args.issue)]
        return self.columns, issue_links + remote_links


class AddLink(JiraLinkMixin, base.JiraShow):
    """Adds new link."""

    def get_parser(self, prog_name):
        parser = super(AddLink, self).get_parser(prog_name)
        parser.add_argument("--type", type=base.utf8, help="Link type", required=True)
        parser.add_argument("--target", type=base.utf8, help="Link target, web link or issue ID", required=True)
        parser.add_argument("--text", type=base.utf8, help="Additional text for link")
        parser.add_argument("--icon", type=base.utf8, help="Icon for link, actual only for remote links")

        return parser

    def take_action(self, parsed_args):
        """Adds new link."""
        if parsed_args.type == self.TYPE_REMOTE_LINK:
            data = {
                "url": parsed_args.target,
                "title": parsed_args.text or self._get_title(parsed_args.target),
                "icon": {"url16x16": parsed_args.icon or self._get_favicon(parsed_args.target)},
            }
            link_id = self.app.jira.add_simple_link(parsed_args.issue, data).id
            link = self.format_remote_link(self.app.jira.remote_link(parsed_args.issue, link_id))
        else:
            link_id = self.app.jira.create_issue_link(
                parsed_args.type, parsed_args.target, parsed_args.issue, parsed_args.text
            ).id
            link = self.format_issue_link(self.app.jira.issue_link(parsed_args.issue, link_id))

        return self.columns, link

    @staticmethod
    def _get_title(url):
        data = urlparse.urlparse(url)
        return urlparse.urlunparse((data[0], data[1], data[2], "", "", ""))

    @staticmethod
    def _get_favicon(url):
        data = urlparse.urlparse(url)
        return urlparse.urlunparse((data[0], data[1], "favicon.ico", "", "", ""))


class ShowLink(JiraLinkMixin, base.JiraShow):
    """Shows a link."""

    def get_parser(self, prog_name):
        parser = super(ShowLink, self).get_parser(prog_name)
        parser.add_argument("--id", type=base.utf8, help="Link ID", required=True)
        return parser

    def take_action(self, parsed_args):
        """Get issue by id."""
        if parsed_args.id.startswith("L"):
            data = self.format_remote_link(self.app.jira.remote_link(parsed_args.issue, parsed_args.id[1:]))
        elif parsed_args.id.startswith("I"):
            data = self.format_issue_link(self.app.jira.issue_link(parsed_args.id[1:]))
        else:
            raise ValueError("Link is not found")
        return self.columns, data


class DropLink(JiraLinkMixin, base.JiraCommand):
    def get_parser(self, prog_name):
        parser = super(DropLink, self).get_parser(prog_name)
        parser.add_argument("--id", type=base.utf8, help="Link ID", required=True)
        return parser

    def take_action(self, parsed_args):
        """Get issue by id."""
        if parsed_args.id.startswith("L"):
            self.app.jira.remote_link(parsed_args.issue, parsed_args.id[1:]).delete()
        elif parsed_args.id.startswith("I"):
            self.app.jira.delete_issue_link(parsed_args.id[1:])
        else:
            raise ValueError("Link is not found")
        self.app.stdout.write("Done.\n")



"""
This file is part of jiractl

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from jiractl.commands import base


class JiraCommentMixin(object):
    columns = ("id", "updated", "author", "text")

    def get_parser(self, prog_name):
        parser = super(JiraCommentMixin, self).get_parser(prog_name)
        parser.add_argument("--issue", type=base.utf8, help="Issue ID", required=True)
        return parser

    @staticmethod
    def format_comment(comment):
        """Converts issue comment to tuple."""
        return comment.id, comment.updated,  comment.author.name, comment.body

    @staticmethod
    def format_text(text):
        """Formats text."""
        return text.replace("<br>", "\n")


class ListComments(JiraCommentMixin, base.JiraList):
    """Gets all comments for issue."""

    def take_action(self, parsed_args):
        """Get issue by id."""
        return self.columns, [self.format_comment(x) for x in self.app.jira.comments(parsed_args.issue)]


class AddComment(JiraCommentMixin, base.JiraShow):
    """Adds new comment."""

    def get_parser(self, prog_name):
        parser = super(AddComment, self).get_parser(prog_name)
        parser.add_argument("--text", type=base.utf8, help="Comment", required=True)
        parser.add_argument(
            "--visibility", type=base.utf8, metavar="TYPE:VALUE", help="Visibility of issue. format %(metavar)"
        )
        return parser

    def take_action(self, parsed_args):
        """Adds new comment."""
        visibility = None
        if parsed_args.visibility:
            visibility = dict(zip(("type", "value"), parsed_args.visibility.split(":", 2)))

        comment = self.app.jira.add_comment(
            parsed_args.issue, self.format_text(parsed_args.text), visibility=visibility
        )
        return self.columns, self.format_comment(comment)


class EditComment(JiraCommentMixin, base.JiraCommand):
    """Edit comment."""

    def get_parser(self, prog_name):
        parser = super(EditComment, self).get_parser(prog_name)
        parser.add_argument("--id", type=base.utf8, help="Comment ID", required=True)
        parser.add_argument("--text", type=base.utf8, help="Comment", required=True)
        parser.add_argument("--visibility", type=base.utf8, help="Visibility in format type:value")
        return parser

    def take_action(self, parsed_args):
        """Get issue by id."""
        visibility = None
        if parsed_args.visibility:
            visibility = dict(zip(("type", "value"), parsed_args.visibility.split(":", 2)))

        comment = self.app.jira.comment(parsed_args.issue, parsed_args.id)
        comment.update(body=self.format_text(parsed_args.text), visibility=visibility)
        self.app.stdout.write("Done.\n")


class ShowComment(JiraCommentMixin, base.JiraShow):
    """Shows one comment."""

    def get_parser(self, prog_name):
        parser = super(ShowComment, self).get_parser(prog_name)
        parser.add_argument("--id", type=base.utf8, help="Comment ID", required=True)
        return parser

    def take_action(self, parsed_args):
        """Get issue by id."""
        return self.columns, self.format_comment(self.app.jira.comment(parsed_args.issue, parsed_args.id))



"""
This file is part of jiractl

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from cliff import app
from cliff.commandmanager import CommandManager
import jira

import jiractl


class JiraApp(app.App):
    """Main cliff application class.

    Performs initialization of the command manager and
    configuration of basic engines.
    """

    _jira = None

    def build_option_parser(self, description, version, argparse_kwargs=None):
        """Specifies global options."""
        parser = super(JiraApp, self).build_option_parser(
            description=description, version=version, argparse_kwargs=argparse_kwargs
        )

        parser.add_argument(
            "-s", "--server",
            metavar="URL",
            help="The Jira URL",
        )
        parser.add_argument(
            "-u", "--user",
            metavar="USERNAME",
            help="Jira user"
        )
        parser.add_argument(
            "-p", "--password",
            metavar="PASSWORD",
            help="Jira password"
        )
        return parser

    @property
    def jira(self):
        if self._jira is None:
            self._jira = jira.JIRA(server=self.options.server, basic_auth=(self.options.user, self.options.password))
        return self._jira


def run_app(cmd_mgr, argv=None):
    """Runs application."""
    return JiraApp(
        description="The command line interface for Jira REST API",
        version=jiractl.__version__,
        command_manager=cmd_mgr,
        deferred_help=True
    ).run(argv)


def main(argv=None):
    """Entry point."""
    return run_app(CommandManager(__package__, convert_underscores=True), argv)


def debug(name, cmd_class, argv=None):
    """Helper for debugging single command without package installation."""
    import sys

    if argv is None:
        argv = sys.argv[1:]

    argv = [name] + argv + ["-v", "-v", "--debug"]
    cmd_mgr = CommandManager(__package__, convert_underscores=True)
    cmd_mgr.add_command(name, cmd_class)

    return run_app(cmd_mgr, argv)



"""
This file is part of jiractl

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import mock

from jiractl.commands import links

from tests import base


class TestListLinks(base.BaseUnitTest):
    command = links.ListLinks

    def test_success(self):
        self.jira.issue.return_value = mock.Mock(fields=mock.Mock(issuelinks=[mock.Mock()]))
        self.jira.remote_links.return_value = [mock.Mock()]

        self.check_output_many(self.command, ['--issue=ISSUE'], 2)

        self.jira.issue.assert_called_once_with("ISSUE")
        self.jira.remote_links.assert_called_once_with("ISSUE")

    def test_required_arguments(self):
        self.check_required_arguments(self.command, ['--issue=ISSUE'])


class TestAddLink(base.BaseUnitTest):
    command = links.AddLink

    def test_add_issue_link_with_text(self):
        self.jira.create_issue.return_value = mock.Mock()
        argv = ['--issue=ISSUE', '--type=DEPENDS', '--target=TARGET', '--text=TEXT']
        self.check_output_one(self.command, argv)
        self.jira.create_issue_link.assert_called_once_with("DEPENDS", "TARGET", "ISSUE", "TEXT")

    def test_add_issue_link_wo_text(self):
        self.jira.create_issue.return_value = mock.Mock()
        argv = ['--issue=ISSUE', '--type=DEPENDS', '--target=TARGET']
        self.check_output_one(self.command, argv)
        self.jira.create_issue_link.assert_called_once_with("DEPENDS", "TARGET", "ISSUE", None)

    def test_add_remote_link_with_text_and_icon(self):
        self.jira.create_issue.return_value = mock.Mock()
        argv = ['--issue=ISSUE', '--type=link', '--target=TARGET', '--text=TEXT', '--icon=ICON']
        self.check_output_one(self.command, argv)
        self.jira.add_simple_link.assert_called_once_with(
            "ISSUE", {"url": "TARGET", "title": "TEXT", "icon": {"url16x16": "ICON"}}
        )

    def test_add_remote_link_wo_text_and_icon(self):
        self.jira.create_issue.return_value = mock.Mock()
        argv = ['--issue=ISSUE', '--type=link', '--target=http://localhost/path?p=v']
        self.check_output_one(self.command, argv)
        self.jira.add_simple_link.assert_called_once_with(
            "ISSUE",
            {
                "url": "http://localhost/path?p=v",
                "title": "http://localhost/path",
                "icon": {"url16x16": "http://localhost/favicon.ico"}
            }
        )

    def test_required_arguments(self):
        argv = ['--issue=ISSUE', '--type=link', '--target=TARGET']
        self.check_required_arguments(self.command, argv)


class TestShowLink(base.BaseUnitTest):
    command = links.ShowLink

    def test_show_issue_link(self):
        argv = ['--issue=ISSUE', '--id=I1']
        self.check_output_one(self.command, argv)
        self.jira.issue_link.assert_called_once_with("1")
        self.assertEqual(0, self.jira.remote_link.call_count)

    def test_show_remote_link(self):
        argv = ['--issue=ISSUE', '--id=L1']
        self.check_output_one(self.command, argv)
        self.jira.remote_link.assert_called_once_with("ISSUE", "1")
        self.assertEqual(0, self.jira.issue_link.call_count)

    def test_show_fails_if_invalid_id(self):
        argv = ['--issue=ISSUE', '--id=1']
        self.check_stderr(self.command, argv, ValueError, "Link is not found")
        self.assertEqual(0, self.jira.remote_link.call_count)
        self.assertEqual(0, self.jira.issue_link.call_count)

    def test_required_arguments(self):
        argv = ['--issue=ISSUE', '--id=L1']
        self.check_required_arguments(self.command, argv)


class TestDropLink(base.BaseUnitTest):
    command = links.DropLink

    def test_delete_issue_link(self):
        argv = ['--issue=ISSUE', '--id=I1']
        self.check_stdout(self.command, argv, "Done.\n")
        self.jira.delete_issue_link.assert_called_once_with("1")
        self.assertEqual(0, self.jira.remote_link.call_count)

    def test_delete_remote_link(self):
        argv = ['--issue=ISSUE', '--id=L1']
        self.check_stdout(self.command, argv, "Done.\n")
        self.jira.remote_link.assert_called_once_with("ISSUE", "1")
        self.jira.remote_link.return_value.delete.assert_called_once_with()
        self.assertEqual(0, self.jira.delete_issue_link.call_count)

    def test_show_fails_if_invalid_id(self):
        argv = ['--issue=ISSUE', '--id=1']
        self.check_stderr(self.command, argv, ValueError, "Link is not found")
        self.assertEqual(0, self.jira.remote_link.call_count)
        self.assertEqual(0, self.jira.delete_issue_link.call_count)

    def test_required_arguments(self):
        argv = ['--issue=ISSUE', '--id=ID']
        self.check_required_arguments(self.command, argv)



"""
This file is part of jiractl

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from jiractl.commands import base


class JiraIssueMixin(object):
    columns = ("project", "id", "key", "summary", "type", "assignee", "status")

    @staticmethod
    def format_issue(issue, custom_fields=None):
        """Converts issue to tuple."""

        result = (
            issue.fields.project.name, issue.id, issue.key, issue.fields.summary,
            issue.fields.issuetype.name, issue.fields.assignee.name, issue.fields.status.name,
        )
        if custom_fields:
            result += tuple(getattr(issue.fields, x) for x in custom_fields)
        return result


class CreateIssue(JiraIssueMixin, base.JiraShow):
    """Creates issue"""

    def get_parser(self, prog_name):
        parser = super(CreateIssue, self).get_parser(prog_name)
        parser.add_argument("--project", type=base.utf8, help="Jira project", required=True)
        parser.add_argument("--type", type=base.utf8, help="Issue Type", required=True)
        parser.add_argument("--summary", type=base.utf8, help="Issue Summary", required=True)
        parser.add_argument("--description", type=base.utf8, help="Issue description", required=True)
        parser.add_argument("--assignee", type=base.utf8, help="Issue Assignee")
        parser.add_argument("--parent", type=base.utf8, help="Issue parent")
        parser.add_argument("--components", type=base.utf8, nargs='+', help="Issue component")
        parser.add_argument("--labels", type=base.utf8, nargs='+', help="Issue label")

        return parser

    def take_action(self, parsed_args):
        """Creates a new jira issue."""
        fields = {
            "project": parsed_args.project,
            "issuetype": parsed_args.type,
            "summary": parsed_args.summary,
            "description": parsed_args.description
        }
        if parsed_args.assignee:
            fields["assignee"] = parsed_args.assignee
        if parsed_args.labels:
            fields["assignee"] = {"name": parsed_args.assignee}
        if parsed_args.parent:
            fields["parent"] = {"id": parsed_args.parent}
        if parsed_args.components:
            fields["components"] = [{'name': c} for c in parsed_args.components]
        if parsed_args.labels:
            fields["labels"] = parsed_args.labels

        return self.columns, self.format_issue(self.app.jira.create_issue(fields=fields, prefetch=True))


class EditIssue(JiraIssueMixin, base.JiraCommand):
    """Updates issue"""

    def get_parser(self, prog_name):
        parser = super(EditIssue, self).get_parser(prog_name)
        parser.add_argument("--id", type=base.utf8, help="Issue ID", required=True)
        parser.add_argument("--summary", type=base.utf8, help="New issue Summary")
        parser.add_argument("--description", type=base.utf8, help="New issue description")
        parser.add_argument("--assignee", type=base.utf8, help="New issue assignee")
        parser.add_argument("--status", type=base.utf8, help="Transition issue to status")
        parser.add_argument("--fields", nargs='+', metavar="NAME:VALUE", type=base.utf8, help="Custom fields to update")

        return parser

    def take_action(self, parsed_args):
        """Update issue."""
        fields = {}
        if parsed_args.summary:
            fields["summary"] = parsed_args.summary
        if parsed_args.description:
            fields["description"] = parsed_args.description
        if parsed_args.fields:
            fields.update((x.split(':', 2)) for x in parsed_args.fields)

        if fields:
            self.app.jira.issue(parsed_args.id).update(fields=fields)

        if parsed_args.assignee:
            self.app.jira.assign_issue(parsed_args.id, parsed_args.assignee)

        if parsed_args.status:
            self.app.jira.transition_issue(parsed_args.id, parsed_args.status)

        self.app.stdout.write("Done.\n")


class ShowIssue(JiraIssueMixin, base.JiraShow):
    """Updates issue"""

    def get_parser(self, prog_name):
        parser = super(ShowIssue, self).get_parser(prog_name)
        parser.add_argument("--id", type=base.utf8, help="Issue ID", required=True)
        return parser

    def take_action(self, parsed_args):
        """Get issue by id."""
        known_fields = set(self.columns)
        custom_fields = tuple(x for x in parsed_args.columns if x not in known_fields)
        columns = self.columns + custom_fields
        return columns, self.format_issue(self.app.jira.issue(parsed_args.id), custom_fields)


class ListIssues(JiraIssueMixin, base.JiraList):
    """Get list of issues which match specified criteria."""

    def get_parser(self, prog_name):
        parser = super(ListIssues, self).get_parser(prog_name)
        parser.add_argument("--project", type=base.utf8, help="Jira project", required=True)
        parser.add_argument("--assignee", type=base.utf8, help="Issue assignee", required=True)
        parser.add_argument("--status", type=base.utf8, nargs='+', help="Issue status", required=True)
        return parser

    def take_action(self, parsed_args):
        """Gets issues."""
        query = u'project="{0}" AND assignee="{1}" AND status IN ("{2}")'.format(
            parsed_args.project, parsed_args.assignee, u'","'.join(x for x in parsed_args.status)
        )
        issues = self.app.jira.search_issues(query)
        return self.columns, [self.format_issue(issue) for issue in issues]


class SearchIssues(JiraIssueMixin, base.JiraList):
    """Searches issues by Jira Query string."""
    def get_parser(self, prog_name):
        parser = super(SearchIssues, self).get_parser(prog_name)
        parser.add_argument("--query", type=base.utf8, help="Query string", required=True)
        return parser

    def take_action(self, parsed_args):
        """Searches issues."""
        issues = self.app.jira.search_issues(parsed_args.query)
        return self.columns, [self.format_issue(issue) for issue in issues]



import setuptools

# In python < 2.7.4, a lazy loading of package `pbr` will break
# setuptools if some other modules registered functions in `atexit`.
# solution from: http://bugs.python.org/issue15881#msg170215
try:
    import multiprocessing  # noqa
except ImportError:
    pass


setuptools.setup(
    setup_requires=['pbr'],
    pbr=True)
